/**
 * Simulation of memory management algorithms
 *
 * Ron Mak
 * Department of Computer Engineering
 * San Jose State University
 */
#include "AllocationAlgorithm.h"

#include <iostream>
#include <stdio.h>
#include "globals.h"
#include "LiveMap.h"

//#define DEBUG
//#define PAUSE

using namespace std;

AllocationAlgorithm::AllocationAlgorithm(string name)
    : name(name), request_count(0), allocate_request_count(0),
      deallocate_request_count(0), onhold_count(0), total_wait(0)
{
    segment_list.push_back(new Segment(-MAX_BLOCKS, 0));
#ifndef DEBUG
    live_map.init();
#endif
}

//void AllocationAlgorithm::add_requester(char name, Semaphore *sem)
//{
//    requester_vector.push_back(name);
//    semaphore_map[name] = sem;
//}

bool AllocationAlgorithm::allocate(const string id, const int size)
{
    request_count++;
    allocate_request_count++;

    int n = try_to_allocate(id, size);

    // If couldn't allocate, put the request on hold.
    if (n < 0) put_request_on_hold(id, size);

    // Memory is full if these vector sizes are equal.
    bool memory_full = onhold_vector.size() == requester_vector.size();
    if (memory_full) process_allocations_on_hold();

    return memory_full;
}

int AllocationAlgorithm::try_to_allocate(const string id, const int size)
{
    // Look for a hole.
    if (!find_hole(id, size)) return -1;

    int n = insert_into_segment_list(id, size);
    memory_map[id] = new Segment(id, size, n);

#ifndef DEBUG
    live_map.show_allocation(id[0], size, n);

#ifdef PAUSE
    live_map.pause();
#endif

#else
    cout << "Allocated " << id << " size " << size << " at " << n << endl;
    print_segment_list();
#endif

    Semaphore *sem = get_requester_semaphore(id);
    SIGNAL(sem);

#ifdef DEBUG
    cout << "Signalled " << sem->name << " id " << id << " size " << size
         << " at " << n << endl;
#endif

    return n;
}

void AllocationAlgorithm::deallocate(const string id)
{
    Segment *segment = memory_map.at(id);

    request_count++;
    deallocate_request_count++;

    memory_map.erase(id);
    remove_from_segment_list(id);

#ifndef DEBUG
    live_map.show_deallocation(segment->size, segment->n);

#ifdef PAUSE
    live_map.pause();
#endif

#else
    cout << "Deallocated " << id << " size " << segment->size
         << " at " << segment->n << endl;
    print_segment_list();
#endif

    process_allocations_on_hold();
}

Semaphore *AllocationAlgorithm::get_requester_semaphore(const string id)
{
    char name = id[0];

    // Is the requester's semaphore already open?
    if (semaphore_map.find(name) != semaphore_map.end())
    {
        // Yes. Return it.
        return semaphore_map[name];
    }
    else
    {
        // No. Open the semaphore for the requester
        //     and enter it into the semaphore map.
        Semaphore *sem = new Semaphore();
        int base_length = strlen(SEMAPHORE_NAME_BASE);
        char *semaphore_name = (char *)malloc(base_length + 2);
        sprintf(semaphore_name, "%s%c", SEMAPHORE_NAME_BASE, name);
        OPEN(sem, semaphore_name);
        semaphore_map[name] = sem;

        requester_vector.push_back(name);
        return sem;
    }
}

void AllocationAlgorithm::put_request_on_hold(const string id, const int size)
{
    OnHold onhold(id[0], id, size, request_count);
    onhold_vector.push_back(onhold);
    onhold_count++;
}

void AllocationAlgorithm::process_allocations_on_hold()
{
    int i = 0;
    while (i < onhold_vector.size())
    {
        OnHold onhold = onhold_vector[i];
        int n = try_to_allocate(onhold.id, onhold.size);

        if (n >= 0)
        {
            // Successful allocation.
            total_wait += (request_count - onhold.request_index);
            onhold_vector.erase(onhold_vector.begin() + i);
        }
        else i++;
    }
}

int AllocationAlgorithm::insert_into_segment_list(const string id,
                                                  const int size)
{
    Segment *segment = *segment_list_iterator;
    int extra = size + segment->size;
    int n = segment->n;

    if (extra == 0)
    {
        // Exact fit.
        segment->id = id;
        segment->size = size;
    }
    else
    {
        // The hole is larger than needed:
        // Insert the new allocation behind the hole
        // and shrink the hole.
        segment_list.insert(segment_list_iterator,
                            new Segment(id, size, n));
        Segment *hole = *segment_list_iterator;
        hole->size = extra;
        hole->n = hole->n + size;
    }

    return n;
}

void AllocationAlgorithm::remove_from_segment_list(const string id)
{
    for (auto it = segment_list.begin(); it != segment_list.end(); it++)
    {
        Segment *segment = *it;

        // If found the segment to remove ...
        if (id == segment->id)
        {
            // ... convert it to a hole.
            segment->id = "";
            segment->size = -segment->size;

            // Merge it with an adjacent hole on the left?
            if (it != segment_list.begin())
            {
                auto prev_it = it;
                Segment *prev_segment = *(--prev_it);

                if (prev_segment->size < 0)
                {
                    if (segment_list_iterator == prev_it)
                    {
                        segment_list_iterator = it;
                    }

                    segment->size += prev_segment->size;
                    segment->n = prev_segment->n;
                    segment_list.erase(prev_it);
                }
            }

            // Merge it with an adjacent hole on the right?
            if (it != prev(segment_list.end()))
            {
                auto next_it = it;
                Segment *next_segment = *(++next_it);

                if (next_segment->size < 0)
                {
                    if (segment_list_iterator == next_it)
                    {
                        segment_list_iterator = it;
                    }

                    segment->size += next_segment->size;
                    segment_list.erase(next_it);
                }
            }

            break;
        }
    }
}

void AllocationAlgorithm::print_segment_list()
{
    for (Segment *segment : segment_list)
    {
        string id = segment->id;

        printf("(");
        if (id.length() > 0) printf("%s ", id.c_str());
        printf("%d %d)\n", segment->size, segment->n);
    }

    fflush(stdout);
}

void AllocationAlgorithm::finish()
{
    request_count++;
#ifndef DEBUG
    live_map.finish();
#endif
}

void AllocationAlgorithm::print_statistics()
{
    char text[80];

    sprintf(text, "%s Statistics", name.c_str());
    live_map.print(text);
    live_map.print();

    sprintf(text, "     Total requests = %4d", request_count);
    live_map.print(text);
    sprintf(text, "  Allocate requests = %4d = %4.1f%%",
            allocate_request_count,
            (100.0*allocate_request_count)/request_count);
    live_map.print(text);
    sprintf(text, "Deallocate requests = %4d = %4.1f%%",
            deallocate_request_count,
            (100.0*deallocate_request_count)/request_count);
    live_map.print(text);

    int still_on_hold_count     = onhold_vector.size();
    int processed_on_hold_count = onhold_count - still_on_hold_count;

    live_map.print();
    sprintf(text, "      Total on hold = %4d", onhold_count);
    live_map.print(text);
    sprintf(text, "  Processed on hold = %4d", processed_on_hold_count);
    live_map.print(text);

    if (processed_on_hold_count > 0)
    {
        sprintf(text, "Processed hold time = %4d", total_wait);
        live_map.print(text);
        sprintf(text, "  Average hold time = %4.1f",
                ((double)total_wait)/processed_on_hold_count);
        live_map.print(text);
    }

    live_map.print();

    int allocated_blocks = 0;
    int free_blocks = 0;

    for (Segment *segment : segment_list)
    {
        int size = segment->size;
        if (size > 0) allocated_blocks += size;
        if (size < 0) free_blocks += size;
    }

    free_blocks = -free_blocks;

    sprintf(text, "   Allocated blocks = %4d = %4.1f%%",
            allocated_blocks, (100.0*allocated_blocks)/MAX_BLOCKS);
    live_map.print(text);
    sprintf(text, "        Free blocks = %4d = %4.1f%%",
            free_blocks, (100.0*free_blocks)/MAX_BLOCKS);
    live_map.print(text);
    live_map.print();
}
